
# JPmag.net
# Multi-language support added by kishe.com
# 0.1 API update by Drapiece (drapiee@gmail.com)

import countrycode


#http://api.tmdb.org/3/movie/276744s/images?api_key=a3dc111e66105f6387e99393813ae4d5

BASE_URL = 'http://jpmag.net/api/plex?' 
JPMAGNET_CONFIG = '%s/configuration' % BASE_URL
JPMAGNET_LOGIN = '%smagtype=userlogin&userid=%%s&password=%%s' % BASE_URL

# Movies

JPMAGNET_MOVIE_SEARCH = '%stoken=%%s&magtype=movie&query=%%s&year=%%s&language=%%s&include_adult=%%s' % BASE_URL
JPMAGNET_MOVIE = '%smagtype=movieinfo&query=%%s&append_to_response=releases,credits&language=%%s&token=%%s' % BASE_URL
JPMAGNET_MOVIE_IMAGES = '%smagtype=movieimage&query=%%s&token=%%s' % BASE_URL

# TV 
JPMAGNET_TV_SEARCH = '%s/search/tv?query=%%s&year=%%s&language=%%s&include_adult=%%s' % BASE_URL
JPMAGNET_TV = '%s/tv/%%s?append_to_response=credits&language=%%s' % BASE_URL
JPMAGNET_TV_SEASON = '%s/tv/%%s/season/%%s?language=%%s' % BASE_URL
JPMAGNET_TV_EPISODE = '%s/tv/%%s/season/%%s/episode/%%s?append_to_response=credits,images&language=%%s' % BASE_URL
JPMAGNET_TV_IMAGES = '%s/tv/%%s/images' % BASE_URL
JPMAGNET_TV_EXTERNAL_IDS = '%s/tv/%%s/external_ids' % BASE_URL
JPMAGNET_TV_TVDB = '%s/tv/find/%%s?external_source=tvdb_id' % BASE_URL

ARTWORK_ITEM_LIMIT = 15
POSTER_SCORE_RATIO = .3 # How much weight to give ratings vs. vote counts when picking best posters. 0 means use only ratings.
BACKDROP_SCORE_RATIO = .3
STILLS_SCORE_RATIO = .3
RE_IMDB_ID = Regex('^tt\d{7}$')

# JPMAGNET does not seem to have an official set of supported languages.  Users can register and 'translate'
# any movie to any ISO 639-1 language.  The following is a realistic list from a popular title.
# This agent falls back to English metadata and sorts out foreign artwork to to ensure the best
# user experience when less common languages are chosen.
LANGUAGES = [
             Locale.Language.English, Locale.Language.Korean, Locale.Language.Japanese
            ]

####################################################################################################
def Start():

  pass

####################################################################################################
@expose
def GetImdbId(jpmagnet_id, lang='en'):

  jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE % (jpmagnet_id, lang))

  if isinstance(jpmagnet_dict, dict) and 'imdb_id' in jpmagnet_dict and RE_IMDB_ID.search(jpmagnet_dict['imdb_id']):
    return jpmagnet_dict['imdb_id']

  return None

####################################################################################################
@expose
def GetTvdbId(jpmagnet_id):

  jpmagnet_dict = GetJSON(url=JPMAGNET_TV_EXTERNAL_IDS % jpmagnet_id)

  if isinstance(jpmagnet_dict, dict) and 'tvdb_id' in jpmagnet_dict and jpmagnet_dict['tvdb_id']:
    return str(jpmagnet_dict['tvdb_id'])

  return None

####################################################################################################
@expose
def GetTvRageId(jpmagnet_id):

  jpmagnet_dict = GetJSON(url=JPMAGNET_TV_EXTERNAL_IDS % jpmagnet_id)

  if isinstance(jpmagnet_dict, dict) and 'tvrage_id' in jpmagnet_dict and jpmagnet_dict['tvrage_id']:
    return str(jpmagnet_dict['tvrage_id'])

  return None

####################################################################################################
@expose
def GetJPmagnetSearchResults(id, name, year, lang, manual, get_imdb_id):

  media = FakeMediaObj(id, name, year)
  results = []
  PerformJPmagnetMovieSearch('', results, media, lang, manual, get_imdb_id)

  return results if len(results) > 0 else None

####################################################################################################
@expose
def GetJPmagnetMetadata(id, lang):

  return PerformJPmagnetMovieUpdate(id, lang)

####################################################################################################
def GetJSON(url, cache_time=0):

  jpmagnet_dict = None

  try:
    jpmagnet_dict = JSON.ObjectFromURL(url, sleep=2.0, headers={'Accept': 'application/json'}, cacheTime=cache_time)
  except:
    Log('Error fetching JSON from JPmag.net.')

  return jpmagnet_dict

####################################################################################################
def AppendSearchResult(results, id, name=None, year=-1, score=0, lang=None):

  new_result = dict(id=str(id), name=name, year=int(year), score=score, lang=lang)

  if isinstance(results, list):

    results.append(new_result)

  else:

    results.Append(MetadataSearchResult(**new_result))

####################################################################################################
def DictToMovieMetadataObj(metadata_dict, metadata):

  try:
    if not metadata or not metadata.attrs:
      return
  except AttributeError:
    Log('WARNING: Framework not new enough to use One True Agent')  # TODO: add a more official log message about version number when available
    return

  Log(metadata)

  for attr_name, attr_obj in metadata.attrs.iteritems():

    if attr_name not in metadata_dict:
      continue
    dict_value = metadata_dict[attr_name]

    if isinstance(dict_value, list):

      attr_obj.clear()
      for val in dict_value:
        attr_obj.add(val)

    elif isinstance(dict_value, dict):

      if attr_name in ['posters', 'art', 'themes', 'directors']:  # Can't access MapObject, so have to write these out

        for k, v in dict_value.iteritems():
          if isinstance(v, tuple):
            attr_obj[k] = Proxy.Preview(HTTP.Request(v[0]).content, sort_order=v[1])
          else:
            attr_obj[k] = v

        attr_obj.validate_keys(dict_value.keys())

      else:
        for k, v in dict_value.iteritems():
          attr_obj[k] = v

    elif attr_name is 'originally_available_at':

      try:
        attr_obj.setcontent(Datetime.ParseDate(dict_value).date())
      except:
        pass

    else:
      attr_obj.setcontent(dict_value)

  metadata.content_rating=metadata_dict['content_ration']
  metadata.tagline=metadata_dict['tagline']
  metadata.title_sort=metadata_dict['title_sort']

  if 'directors' in metadata_dict:
    metadata.directors.clear()
    for director in metadata_dict['directors']:
      meta_direc = metadata.directors.new()
      if 'name' in director:
        meta_direc.name = director['name']

  # Roles is a special kind of object
  if 'roles' in metadata_dict:
    metadata.roles.clear()

    for role in metadata_dict['roles']:
      meta_role = metadata.roles.new()
      if 'role' in role:
        meta_role.role = role['role']

      if 'actor' in role:
        meta_role.name = role['actor']

      if 'photo' in role:
        meta_role.photo = role['photo']

####################################################################################################
def PerformJPmagnetMovieSearch(token, results, media, lang, manual, get_imdb_id=False):

  # If search is initiated by a different, primary metadata agent.
  # This requires the other agent to use the IMDb id as key.
  if media.primary_metadata is not None and RE_IMDB_ID.search(media.primary_metadata.id):
    AppendSearchResult(results=results, id=media.primary_metadata.id, score=100)

  else:
    # If this a manual search (Fix Incorrect Match) and we get an IMDb id as input.
    if manual and RE_IMDB_ID.search(media.name):
      jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE % (media.name, lang, token))

      if isinstance(jpmagnet_dict, dict) and 'id' in jpmagnet_dict:

        if get_imdb_id:
          if 'imdb_id' in jpmagnet_dict and RE_IMDB_ID.search(jpmagnet_dict['imdb_id']):
            id = str(jpmagnet_dict['imdb_id'])
          else:
            id = GetImdbId(jpmagnet_dict['id'], lang) or jpmagnet['id']
        else:
          id = jpmagnet_dict['id']

        AppendSearchResult(results=results,
                           id=id,
                           name=jpmagnet_dict['title'],
                           year=int(jpmagnet_dict['release_date'].split('-')[0]),
                           score=100,
                           lang=lang)

    # If this is an automatic search and JPmag.net agent is used as a primary agent.
    else:
      if media.year and int(media.year) > 1900:
        year = media.year
      else:
        year = ''

      include_adult = 'false'
      if Prefs['adult']:
        include_adult = 'true'

      # Historically we've StrippedDiacritics() here, but this is a pretty aggressive function that won't pass
      # anything that can't be encoded to ASCII, and as such has a tendency to nuke whole titles in, e.g., Asian
      # languages (See GHI #26).  If we have a string that was modified by StripDiacritics() and we get no results,
      # try the search again with the original.
      #
      stripped_name = String.StripDiacritics(media.name)
      jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE_SEARCH % (String.Quote(token),String.Quote(stripped_name), year, lang, include_adult))
      if media.name != stripped_name and (jpmagnet_dict == None or len(jpmagnet_dict['results']) == 0):
        Log('No results for title modified by strip diacritics, searching again with the original: ' + media.name)
        jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE_SEARCH % (String.Quote(token),String.Quote(media.name), year, lang, include_adult))

      if isinstance(jpmagnet_dict, dict) and 'results' in jpmagnet_dict:

        for i, movie in enumerate(sorted(jpmagnet_dict['results'], key=lambda k: k['popularity'], reverse=True)):
          score = 100
          #score = score - abs(String.LevenshteinDistance(movie['title'].lower(), media.name.lower()))

          # Adjust score slightly for 'popularity' (helpful for similar or identical titles when no media.year is present)
          score = score - (5 * i)

          if 'release_date' in movie and movie['release_date']:
            release_year = int(movie['release_date'].split('-')[0])
          else:
            release_year = -1

          if media.year and int(media.year) > 1900 and release_year:
            year_diff = abs(int(media.year) - release_year)

            if year_diff <= 1:
              score = score + 10
            else:
              score = score - (5 * year_diff)

          if score <= 0:
            continue
          else:

            if get_imdb_id:
              if 'imdb_id' in movie and RE_IMDB_ID.search(movie['imdb_id']):
                id = str(movie['imdb_id'])
              else:
                id = GetImdbId(movie['id'], lang) or movie['id']
            else:
              id = movie['id']

            AppendSearchResult(results=results,
                               id=id,
                               name=movie['title'],
                               year=release_year,
                               score=score,
                               lang=lang)

####################################################################################################
def PerformJPmagnetMovieUpdate(metadata_id, lang, token):

  metadata = dict(id=metadata_id)

#  config_dict = GetJSON(url=JPMAGNET_CONFIG, cache_time=CACHE_1WEEK * 2)
  jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE % (metadata_id, lang, token))

  if not isinstance(jpmagnet_dict, dict) or 'overview' not in jpmagnet_dict or jpmagnet_dict['overview'] is None or jpmagnet_dict['overview'] == "":
    # Retry the query with no language specified if we didn't get anything from the initial request.
    jpmagnet_dict = GetJSON(url=JPMAGNET_MOVIE % (metadata_id, lang, token))

  # This additional request is necessary since full art/poster lists are not returned if they don't exactly match the language
  jpmagnet_images_dict = GetJSON(url=JPMAGNET_MOVIE_IMAGES % (metadata_id , token))

  if not isinstance(jpmagnet_dict, dict) or not isinstance(jpmagnet_images_dict, dict):
    return None

  # Rating.
  votes = jpmagnet_dict['vote_count']
  rating = jpmagnet_dict['vote_average']

  Log("JPMAGNET_DICT")
  Log(jpmagnet_dict)

  if votes > 3:
    metadata['rating'] = rating
    metadata['audience_rating'] = 0.0
    metadata['rating_image'] = None
    metadata['audience_rating_image'] = None

  # Title of the film.
  metadata['title'] = jpmagnet_dict['title']
  metadata['title_sort'] = jpmagnet_dict['title_sort']
  Log("MAKE : TITLE")

  if 'original_title' in jpmagnet_dict and jpmagnet_dict['original_title'] != jpmagnet_dict['title']:
    metadata['original_title'] = jpmagnet_dict['original_title']

  # Tagline.
  metadata['tagline'] = jpmagnet_dict['tagline']

  # Release date.
  try:
    metadata['originally_available_at'] = jpmagnet_dict['release_date']
    metadata['year'] = Datetime.ParseDate(jpmagnet_dict['release_date']).date().year
  except:
    pass

  Log("MAKE : COUNTRY")
#  if Prefs['country'] != '':
#    c = Prefs['country']
#    Log(c+' '+countrycode.COUNTRY_TO_CODE[c]+' '+jpmagnet_dict['releases']['countries']['iso_3166_1'])
#    for country in jpmagnet_dict['releases']['countries']:

#      if country['iso_3166_1'] == countrycode.COUNTRY_TO_CODE[c]:

        # Content rating.
#        if 'certification' in country and country['certification'] != '':
#          if countrycode.COUNTRY_TO_CODE[c] == 'US':
#            metadata['content_rating'] = country['certification']
#          else:
#            metadata['content_rating'] = '%s/%s' % (countrycode.COUNTRY_TO_CODE[c].lower(), country['certification'])

        # Release date (country specific).
#        if 'release_date' in country and country['release_date'] != '':
#          try:
#            metadata['originally_available_at'] = country['release_date']
#            metadata['year'] = Datetime.ParseDate(country['release_date']).date().year
#          except:
#            pass
#        break

  Log("MAKE : SUMMARY")
  # Summary.
  metadata['summary'] = jpmagnet_dict['overview']
  
  # Runtime.
  try: metadata['duration'] = int(jpmagnet_dict['runtime']) * 60 * 1000
  except: pass

  Log("MAKE : GENRES")
  # Genres.
  metadata['genres'] = []
  for genre in jpmagnet_dict['genres']:
    metadata['genres'].append(genre['name'].strip())

  Log("MAKE : COLLECTION")
  # Collections.
  metadata['collections'] = []
  if Prefs['collections'] and jpmagnet_dict['belongs_to_collection'] is not None:
    metadata['collections'].append(jpmagnet_dict['belongs_to_collection']['name'].replace(' Collection',''))

  Log("MAKE : STUDIO")
  # Studio.
  if 'production_companies' in jpmagnet_dict and len(jpmagnet_dict['production_companies']) > 0:
    index = jpmagnet_dict['production_companies'][0]['id']
    company = None

    for studio in jpmagnet_dict['production_companies']:
      if studio['id'] <= index:
        index = studio['id']
        company = studio['name'].strip()

    metadata['studio'] = company
 
  else:
    metadata['studio'] = None

  Log("MAKE : COUNTRY")
  # Country.
  metadata['countries'] = []
  if 'production_countries' in jpmagnet_dict:
    for country in jpmagnet_dict['production_countries']:
      countryx = country['name']
      metadata['countries'].append(countryx)

  Log("MAKE : CREW")
  # Crew.
  metadata['writers'] = []
  metadata['producers'] = []
  metadata['directors'] = []


  for member in jpmagnet_dict['credits']['crew']:
    if member['job'] == 'Director':
      Log(member['name'])
      metadata['directors'].append(member)
    elif member['job'] in ('Writer', 'Screenplay', 'Author'):
      metadata['writers'].append(member)
    elif member['job'] == 'Producer':
      if member['name'] !='':
        metadata['producers'].append(member)


  Log("MAKE : ROLES")
  # Cast.
  
  metadata['roles'] = []

  for member in jpmagnet_dict['credits']['cast']:
    role = {}
    role['role'] = member['character']
    role['actor'] = member['name']
    role['photo'] = member['profile_path']
    metadata['roles'].append(role)


  metadata['content_ration'] = jpmagnet_dict['content_ration']
  # Note: for JPMAGNET artwork, number of votes is a good predictor of poster quality. Ratings are assigned
  # using a Baysean average that appears to be poorly calibrated, so ratings are almost always between
  # 5 and 6 or zero.  Consider both of these, weighting them according to the POSTER_SCORE_RATIO.

  # No votes get zero, use JPMAGNET's apparent initial Baysean prior mean of 5 instead.
  Log("MAKE : POSTERS")
  valid_names = list()
  metadata['posters'] = {}
  Log(jpmagnet_images_dict['posters'])

  if jpmagnet_images_dict['posters']:
    max_average = max([(lambda p: p['vote_average'] or 5)(p) for p in jpmagnet_images_dict['posters']])
    max_count = max([(lambda p: p['vote_count'])(p) for p in jpmagnet_images_dict['posters']]) or 1

    for i, poster in enumerate(jpmagnet_images_dict['posters']):

      score = (poster['vote_average'] / max_average) * POSTER_SCORE_RATIO
      score += (poster['vote_count'] / max_count) * (1 - POSTER_SCORE_RATIO)
      jpmagnet_images_dict['posters'][i]['score'] = score

      # Boost the score for localized posters (according to the preference).
      if Prefs['localart']:
        if poster['iso_639_1'] == lang:
          jpmagnet_images_dict['posters'][i]['score'] = poster['score'] + 1

      # Discount score for foreign posters.
      if poster['iso_639_1'] != lang and poster['iso_639_1'] is not None and poster['iso_639_1'] != 'en':
        jpmagnet_images_dict['posters'][i]['score'] = poster['score'] - 1

    for i, poster in enumerate(sorted(jpmagnet_images_dict['posters'], key=lambda k: k['score'], reverse=True)):
      if i > ARTWORK_ITEM_LIMIT:
        break
      else:
        poster_url = poster['file_path']
        thumb_url = poster['file_path']
        valid_names.append(poster_url)

        if poster_url not in metadata['posters']:
          try: metadata['posters'][poster_url] = (thumb_url, i+1)
          except: pass


  Log("MAKE : BACKDROPS")
  # Backdrops.
  valid_names = list()
  metadata['art'] = {}
  if jpmagnet_images_dict['backdrops']:
    max_average = max([(lambda p: p['vote_average'] or 5)(p) for p in jpmagnet_images_dict['backdrops']])
    max_count = max([(lambda p: p['vote_count'])(p) for p in jpmagnet_images_dict['backdrops']]) or 1

    for i, backdrop in enumerate(jpmagnet_images_dict['backdrops']):

      score = (backdrop['vote_average'] / max_average) * BACKDROP_SCORE_RATIO
      score += (backdrop['vote_count'] / max_count) * (1 - BACKDROP_SCORE_RATIO)
      jpmagnet_images_dict['backdrops'][i]['score'] = score

      # For backdrops, we prefer "No Language" since they're intended to sit behind text.
      if backdrop['iso_639_1'] == 'xx' or backdrop['iso_639_1'] == 'none':
        jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) + 2

      # Boost the score for localized art (according to the preference).
      if Prefs['localart']:
        if backdrop['iso_639_1'] == lang:
          jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) + 1

      # Discount score for foreign art.
      if backdrop['iso_639_1'] != lang and backdrop['iso_639_1'] is not None and backdrop['iso_639_1'] != 'en':
        jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) - 1

    Log("ENTER BACKDROP")

    for i, backdrop in enumerate(sorted(jpmagnet_images_dict['backdrops'], key=lambda k: k['score'], reverse=True)):
      if i > ARTWORK_ITEM_LIMIT:
        break
      else:
        backdrop_url = backdrop['file_path']
        thumb_url = backdrop['file_path']
        valid_names.append(backdrop_url)

        if backdrop_url != '':
          if backdrop_url not in metadata['art']:
            try:
              metadata['art'][backdrop_url] = (thumb_url, i+1)
            except: 
              pass

#    metadata.art.validate_keys(valid_names)

  Log("METADATAEND")
  Log(metadata)
  return metadata


def LoginJPMagnet():

#  proxy = XMLRPC.Proxy(JPMAGNET_LOGIN)
  username = Prefs['username'] if Prefs['username'] else ''
  password = Prefs['password'] if Prefs['password'] else ''
  proxy = GetJSON(url=JPMAGNET_LOGIN % (username , password))
  Log(proxy)
  if proxy['api_key']!='':
    API_KEY = proxy['api_key']
    Log(API_KEY)
    return API_KEY
  else:
    return ""

####################################################################################################
class JPmagnetAgent(Agent.Movies):

  name = 'JPmag.net'
  languages = LANGUAGES
  primary_provider = True
  accepts_from = ['com.plexapp.agents.localmedia']
  contributes_to = ['com.plexapp.agents.imdb','com.plexapp.agents.data18']

  def search(self, results, media, lang, manual):

    token = LoginJPMagnet()
    if token != '':
      PerformJPmagnetMovieSearch(token, results, media, lang, manual)
    else: 
      Log('Unable to retrieve valid token. Skipping')


  def update(self, metadata, media, lang):

    token = LoginJPMagnet()
    if token != '':
      metadata_dict = PerformJPmagnetMovieUpdate(metadata.id, lang, token)
      DictToMovieMetadataObj(metadata_dict, metadata)
    else: 
      Log('Unable to retrieve valid token. Skipping')


####################################################################################################
class JPmagnetAgent(Agent.TV_Shows):

  name = 'JPmag.net'
  languages = LANGUAGES
  primary_provider = True
  accepts_from = ['com.plexapp.agents.localmedia', 'com.plexapp.agents.thetvdb']
  contributes_to = ['com.plexapp.agents.thetvdb']

  def search(self, results, media, lang, manual):

    # If JPMAGNET is used as a secondary agent for TVDB, find the JPMAGNET id
    if media.primary_agent == 'com.plexapp.agents.thetvdb':
      jpmagnet_dict = GetJSON(url=JPMAGNET_TV_TVDB % (media.primary_metadata.id))

      if isinstance(jpmagnet_dict, dict) and 'tv_results' in jpmagnet_dict and len(jpmagnet_dict['tv_results']) > 0:
        jpmagnet_id = jpmagnet_dict['tv_results'][0]['id']

        results.Append(MetadataSearchResult(
          id = str(jpmagnet_id),
          score = 100
        ))

      return

    if media.year and int(media.year) > 1900:
      year = media.year
    else:
      year = ''

    include_adult = 'false'
    if Prefs['adult']:
      include_adult = 'true'

    media_show = media.show
    if not manual:
      filename = String.Unquote(media.filename)
      if (' (US)' in filename or ' (UK)' in filename) and (media.show.endswith(' Us') or media.show.endswith(' Uk')):
        media_show = media.show.rsplit(' ', 1)[0]

    # StrippedDiacritics() is a pretty aggressive function that won't pass
    # anything that can't be encoded to ASCII, and as such has a tendency to nuke whole titles in, e.g., Asian
    # languages (See GHI #26).  If we have a string that was modified by StripDiacritics() and we get no results,
    # try the search again with the original.
    #
    stripped_name = String.StripDiacritics(media_show)
    jpmagnet_dict = GetJSON(url=JPMAGNET_TV_SEARCH % (String.Quote(stripped_name), year, lang, include_adult))

    if media_show != stripped_name and (jpmagnet_dict == None or len(jpmagnet_dict['results']) == 0):
      Log('No results for title modified by strip diacritics, searching again with the original: ' + media_show)
      jpmagnet_dict = GetJSON(url=JPMAGNET_TV_SEARCH % (String.Quote(media_show), year, lang, include_adult))

    if isinstance(jpmagnet_dict, dict) and 'results' in jpmagnet_dict:
      for i, show in enumerate(sorted(jpmagnet_dict['results'], key=lambda k: k['popularity'], reverse=True)):
        score = 90
        score = score - abs(String.LevenshteinDistance(show['name'].lower(), media_show.lower()))

        # Adjust score slightly for 'popularity' (helpful for similar or identical titles when no media.year is present)
        #score = score - (5 * i)

        if 'first_air_date' in show and show['first_air_date']:
          release_year = int(show['first_air_date'].split('-')[0])
          score = score + 5
        else:
          release_year = None
          score = score - 5

        if media.year and int(media.year) > 1900 and release_year:
          year_diff = abs(int(media.year) - release_year)

          if year_diff == 0:
            score = score + 10
          else:
            score = score - (5 * year_diff)

        if score <= 0:
          continue
        else:
          results.Append(MetadataSearchResult(
            id = str(show['id']),
            name = show['name'],
            year = release_year,
            score = score,
            lang = lang
          ))

  def update(self, metadata, media, lang):

    config_dict = GetJSON(url=JPMAGNET_CONFIG, cache_time=CACHE_1WEEK * 2)
    jpmagnet_dict = GetJSON(url=JPMAGNET_TV % (metadata.id, lang))

    if not isinstance(jpmagnet_dict, dict) or 'overview' not in jpmagnet_dict or jpmagnet_dict['overview'] is None or jpmagnet_dict['overview'] == "":
      # Retry the query with no language specified if we didn't get anything from the initial request.
      jpmagnet_dict = GetJSON(url=JPMAGNET_TV % (metadata.id, ''))

    # This additional request is necessary since full art/poster lists are not returned if they don't exactly match the language
    jpmagnet_images_dict = GetJSON(url=JPMAGNET_TV_IMAGES % metadata.id)

    if not isinstance(jpmagnet_dict, dict) or not isinstance(jpmagnet_images_dict, dict):
      return None

    # Rating.
    votes = jpmagnet_dict['vote_count']
    rating = jpmagnet_dict['vote_average']
    if votes > 3:
      metadata.rating = rating
      metadata.audience_rating = 0.0

    # Name of the show.
    metadata.title = jpmagnet_dict['name']

    # First air date.
    try:
      metadata.originally_available_at = Datetime.ParseDate(jpmagnet_dict['first_air_date']).date()
      metadata.year = metadata.originally_available_at.year
    except:
      pass

    # Summary.
    metadata.summary = jpmagnet_dict['overview']
    if metadata.summary == 'No overview found.':
      metadata.summary = ""

    # Runtime.
    try: metadata.duration = int(jpmagnet_dict['episode_run_time'][0]) * 60 * 1000
    except: pass

    # Genres.
    metadata.genres.clear()
    for genre in jpmagnet_dict['genres']:
      metadata.genres.add(genre['name'].strip())

    # Studio.
    if 'production_companies' in jpmagnet_dict and len(jpmagnet_dict['production_companies']) > 0:
      index = jpmagnet_dict['production_companies'][0]['id']

      for studio in jpmagnet_dict['production_companies']:
        if studio['id'] <= index:
          index = studio['id']
          company = studio['name'].strip()

      metadata.studio = company

    else:
      metadata.studio = None

    # Country.
    metadata.countries.clear()
    if 'origin_country' in jpmagnet_dict:
      for country in jpmagnet_dict['origin_country']:
        if len(country) == 2 and country in countrycode.CODE_TO_COUNTRY:
          country = countrycode.CODE_TO_COUNTRY[country]
        else:
          continue

        country = country.replace('United States of America', 'USA')
        metadata.countries.add(country)

    # Cast.
    metadata.roles.clear()

    for member in sorted(jpmagnet_dict['credits']['cast'], key=lambda k: k['order']):
      role = metadata.roles.new()
      role.role = member['character']
      role.actor = member['name']
      if member['profile_path'] is not None:
        role.photo = config_dict['images']['base_url'] + 'original' + member['profile_path']

    # Note: for JPMAGNET artwork, number of votes is a good predictor of poster quality. Ratings are assigned
    # using a Baysean average that appears to be poorly calibrated, so ratings are almost always between
    # 5 and 6 or zero.  Consider both of these, weighting them according to the POSTER_SCORE_RATIO.

    # No votes get zero, use JPMAGNET's apparent initial Baysean prior mean of 5 instead.
    valid_names = list()

    if jpmagnet_images_dict['posters']:
      max_average = max([(lambda p: p['vote_average'] or 5)(p) for p in jpmagnet_images_dict['posters']])
      max_count = max([(lambda p: p['vote_count'])(p) for p in jpmagnet_images_dict['posters']]) or 1

      for i, poster in enumerate(jpmagnet_images_dict['posters']):

        score = (poster['vote_average'] / max_average) * POSTER_SCORE_RATIO
        score += (poster['vote_count'] / max_count) * (1 - POSTER_SCORE_RATIO)
        jpmagnet_images_dict['posters'][i]['score'] = score

        # Boost the score for localized posters (according to the preference).
        if Prefs['localart']:
          if poster['iso_639_1'] == lang:
            jpmagnet_images_dict['posters'][i]['score'] = poster['score'] + 1

        # Discount score for foreign posters.
        if poster['iso_639_1'] != lang and poster['iso_639_1'] is not None and poster['iso_639_1'] != 'en':
          jpmagnet_images_dict['posters'][i]['score'] = poster['score'] - 1

      for i, poster in enumerate(sorted(jpmagnet_images_dict['posters'], key=lambda k: k['score'], reverse=True)):
        if i > ARTWORK_ITEM_LIMIT:
          break
        else:
          poster_url = config_dict['images']['base_url'] + 'original' + poster['file_path']
          thumb_url = config_dict['images']['base_url'] + 'w154' + poster['file_path']
          valid_names.append(poster_url)

          if poster_url not in metadata.posters:
            try: metadata.posters[poster_url] = Proxy.Preview(HTTP.Request(thumb_url).content, sort_order=i+1)
            except: pass

    metadata.posters.validate_keys(valid_names)

    # Backdrops.
    valid_names = list()

    if jpmagnet_images_dict['backdrops']:
      max_average = max([(lambda p: p['vote_average'] or 5)(p) for p in jpmagnet_images_dict['backdrops']])
      max_count = max([(lambda p: p['vote_count'])(p) for p in jpmagnet_images_dict['backdrops']]) or 1

      for i, backdrop in enumerate(jpmagnet_images_dict['backdrops']):

        score = (backdrop['vote_average'] / max_average) * BACKDROP_SCORE_RATIO
        score += (backdrop['vote_count'] / max_count) * (1 - BACKDROP_SCORE_RATIO)
        jpmagnet_images_dict['backdrops'][i]['score'] = score

        # For backdrops, we prefer "No Language" since they're intended to sit behind text.
        if backdrop['iso_639_1'] == 'xx' or backdrop['iso_639_1'] == 'none':
          jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) + 2

        # Boost the score for localized art (according to the preference).
        if Prefs['localart']:
          if backdrop['iso_639_1'] == lang:
            jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) + 1

        # Discount score for foreign art.
        if backdrop['iso_639_1'] != lang and backdrop['iso_639_1'] is not None and backdrop['iso_639_1'] != 'en':
          jpmagnet_images_dict['backdrops'][i]['score'] = float(backdrop['score']) - 1

      for i, backdrop in enumerate(sorted(jpmagnet_images_dict['backdrops'], key=lambda k: k['score'], reverse=True)):
        if i > ARTWORK_ITEM_LIMIT:
          break
        else:
          backdrop_url = config_dict['images']['base_url'] + 'original' + backdrop['file_path']
          thumb_url = config_dict['images']['base_url'] + 'w300' + backdrop['file_path']
          valid_names.append(backdrop_url)

          if backdrop_url not in metadata.art:
            try: metadata.art[backdrop_url] = Proxy.Preview(HTTP.Request(thumb_url).content, sort_order=i+1)
            except: pass

    metadata.art.validate_keys(valid_names)

    # Get episode data.
    @parallelize
    def UpdateEpisodes():

      # Loop over seasons.
      for s in media.seasons:
        season = metadata.seasons[s]

        # Set season metadata.
        @task
        def UpdateSeason(season=season, s=s):

          jpmagnet_season_dict = GetJSON(url=JPMAGNET_TV_SEASON % (metadata.id, s, lang))

          if jpmagnet_season_dict is None:
            return None

          season.summary = jpmagnet_season_dict['overview']

          # Season poster.
          valid_names = list()

          if 'poster_path' in jpmagnet_season_dict and jpmagnet_season_dict['poster_path']:
#            poster_url = config_dict['images']['base_url'] + 'original' + jpmagnet_season_dict['poster_path']
#            thumb_url = config_dict['images']['base_url'] + 'w154' + jpmagnet_season_dict['poster_path']
            poster_url = config_dict['images']['base_url'] + 'original' + jpmagnet_season_dict['poster_path']
            thumb_url = config_dict['images']['base_url'] + 'w154' + jpmagnet_season_dict['poster_path']
            valid_names.append(poster_url)

            if poster_url not in season.posters:
              try:
                season.posters[poster_url] = Proxy.Preview(HTTP.Request(thumb_url).content)
              except:
                pass

          season.posters.validate_keys(valid_names)

        # Loop over episodes in a season.
        for e in media.seasons[s].episodes:
          episode = metadata.seasons[s].episodes[e]

          # Set episode metadata.
          @task
          def UpdateEpisode(episode=episode, s=s, e=e):

            jpmagnet_episode_dict = GetJSON(url=JPMAGNET_TV_EPISODE % (metadata.id, s, e, lang))

            if not isinstance(jpmagnet_episode_dict, dict):
              return None

            # Rating.
            votes = jpmagnet_episode_dict['vote_count']
            rating = jpmagnet_episode_dict['vote_average']
            if votes > 3:
              episode.rating = rating

            # Name of the episode.
            episode.title = jpmagnet_episode_dict['name']

            # Summary.
            episode.summary = jpmagnet_episode_dict['overview']

            # Air date.
            try:
              episode.originally_available_at = Datetime.ParseDate(jpmagnet_episode_dict['air_date']).date()
            except:
              pass

            # Crew.
            episode.directors.clear()
            episode.writers.clear()
            episode.producers.clear()

            if 'crew' in jpmagnet_episode_dict['credits']:

              for member in jpmagnet_episode_dict['credits']['crew']:
                if member['job'] == 'Director':
                  episode.directors.add(member)
                elif member['job'] in ('Writer', 'Screenplay'):
                  episode.writers.add(member)
                elif member['job'] == 'Producer':
                  episode.producers.add(member)

            # TODO:
            # - Actors per episode?
            # - Guest stars?

            # Episode still.
            valid_names = list()

            # Check if there are stills present and then loop through them
            if jpmagnet_episode_dict['images']['stills']:
              max_average = max([(lambda p: p['vote_average'] or 5)(p) for p in jpmagnet_episode_dict['images']['stills']])
              max_count = max([(lambda p: p['vote_count'])(p) for p in jpmagnet_episode_dict['images']['stills']]) or 1

              for i, still in enumerate(jpmagnet_episode_dict['images']['stills']):

                score = (still['vote_average'] / max_average) * STILLS_SCORE_RATIO
                score += (still['vote_count'] / max_count) * (1 - STILLS_SCORE_RATIO)
                jpmagnet_episode_dict['images']['stills'][i]['score'] = score

              for i, still in enumerate(sorted(jpmagnet_episode_dict['images']['stills'], key=lambda k: k['score'], reverse=True)):
                if i > ARTWORK_ITEM_LIMIT:
                  break
                else:
                  still_url = config_dict['images']['base_url'] + 'original' + still['file_path']
                  thumb_url = config_dict['images']['base_url'] + 'w154' + still['file_path']
                  valid_names.append(still_url)

                  if still_url not in episode.thumbs:
                    try: episode.thumbs[still_url] = Proxy.Preview(HTTP.Request(thumb_url).content, sort_order=i+1)
                    except: pass

            episode.thumbs.validate_keys(valid_names)

####################################################################################################
class FakeMediaObj():

  def __init__(self, id, name, year):
    self.name = name
    self.year = year
    self.primary_metadata = FakePrimaryMetadataObj(id)

####################################################################################################
class FakePrimaryMetadataObj():

  def __init__(self, id):
    self.id = id
